import React from 'react';
import styles from './inputs.module.scss';
import { DropdownButton, DropdownContainer, DropdownItem, DropdownItemContainer } from '../dropdowns';

type SliderProps = {
  label: string;
  type: 'slider';
  value: number;
  min: number;
  max: number;
  step: number;
  showValue?: boolean;
  unit?: string;
  onChange?: (value: number) => void;
};

type TextProps = {
  label: string;
  type: 'text';
  placeholder?: string;
  value: string;
  required?: boolean;
  errorText?: string;
  visible?: boolean;
  disabled?: boolean;
  validate?: (value: any) => boolean;
  onChange?: (value: string) => void;
};

type CheckboxProps = {
  label?: string;
  type: 'checkbox';
  options: Array<string>;
  value: Array<string>;
  onChange?: (value: Array<string>) => void;
};

type BooleanProps = {
  label: string;
  type: 'boolean';
  options?: any;
  value: boolean;
  onChange?: (value: boolean) => void;
};

type RadioProps = {
  label?: string;
  type: 'radio';
  options: Array<string>;
  value: string;
  onChange?: (value: string) => void;
};

type DropdownProps = {
  label?: string;
  value: string | number;
  type: 'dropdown';
  options: any;
  onChange?: (value: string | number) => void;
  required?: boolean;
};

export type InputProps = TextProps | SliderProps | CheckboxProps | DropdownProps | RadioProps | BooleanProps;

const Input = (props: InputProps) => {
  switch (props.type) {
    case 'slider':
      return <Slider {...(props as SliderProps)} />;
    case 'text' || 'password':
      return <Text {...(props as TextProps)} />;
    case 'checkbox':
      return <Checkbox {...(props as CheckboxProps)} />;
    case 'dropdown':
      return <DropDown {...(props as DropdownProps)} />;
    case 'radio':
      return <Radio {...(props as RadioProps)} />;
    case 'boolean':
      return <Boolean {...(props as BooleanProps)} />;
    default:
      return null;
  }
};

export const Slider = ({ label, value, min, max, step, unit, showValue = true, onChange }: SliderProps) => {
  return (
    <div className={styles['slider']}>
      <label className="label flex flex-row justify-between items-end">
        <span className="label-text">{label}</span>
        {showValue ? (
          <div className="label-text">
            {value}
            {unit}
          </div>
        ) : null}
      </label>

      <input
        type="range"
        min={min}
        max={max}
        step={step}
        value={value}
        onChange={(e) => {
          if (onChange) {
            onChange(parseFloat(e.target.value));
          }
        }}
        aria-labelledby="input-slider"
      />
    </div>
  );
};

export const Text = ({
  label,
  placeholder,
  value = '',
  required = false,
  visible = true,
  errorText,
  validate,
  disabled = false,
  onChange,
}: TextProps) => {
  const [isValid, setIsValid] = React.useState<boolean>(true);

  return (
    <div className="form-control w-full">
      <label className="label">
        <span className={`text-sm font-medium text-secondary-700 ${required && "after:content-['*'] after:ml-0.5 after:text-danger-500"}`}>
          {label}
        </span>
        {required && isValid ? null : <span className="label-text-alt text-error">{errorText}</span>}
      </label>
      <input
        type={visible ? 'text' : 'password'}
        placeholder={placeholder}
        className={`px-3 py-2 bg-light border border-secondary-300 placeholder-secondary-400 focus:outline-none block w-full sm:text-sm focus:ring-1 ${
          isValid ? '' : 'input-error'
        }`}
        value={value.toString()}
        onChange={(e) => {
          if (required && validate) {
            setIsValid(validate(e.target.value));
          }
          if (onChange) {
            onChange(e.target.value);
          }
        }}
        required={required}
        disabled={disabled}
      />
    </div>
  );
};

export const Radio = ({ label, value, options, onChange }: RadioProps) => {
  return (
    <div>
      <label className="label">
        <span className="label-text">{label}</span>
      </label>
      {options.map((option, index) => (
        <label key={index} className="label cursor-pointer w-fit gap-2 px-0 py-1">
          <input
            type="radio"
            name={option}
            checked={value === option}
            onChange={() => {
              if (onChange) {
                onChange(option);
              }
            }}
            className="radio radio-xs radio-primary"
          />
          <span className="label-text">{option}</span>
        </label>
      ))}
    </div>
  );
};

export const Checkbox = ({ label, value, options, onChange }: CheckboxProps) => {
  return (
    <div>
      {label && (
        <label className="label">
          <span className="label-text">{label}</span>
        </label>
      )}
      {options.map((option, index) => (
        <label key={index} className="label cursor-pointer w-fit gap-2 px-0 py-1">
          <input
            type="checkbox"
            name={option}
            checked={Array.isArray(value) && value.includes(option)}
            onChange={(event) => {
              const updatedValue = event.target.checked ? [...value, option] : value.filter((val) => val !== option);
              if (onChange) {
                onChange(updatedValue);
              }
            }}
            className="checkbox checkbox-xs"
          />
          <span className="label-text">{option}</span>
        </label>
      ))}
    </div>
  );
};

export const Boolean = ({ label, value, onChange }: BooleanProps) => {
  return (
    <div>
      <label className="label cursor-pointer w-fit gap-2 px-0 py-1">
        <span className="label-text">{label}</span>
        <input
          type="checkbox"
          checked={value}
          onChange={(event) => {
            if (onChange) {
              onChange(event.target.checked);
            }
          }}
          className="checkbox checkbox-xs"
        />
      </label>
    </div>
  );
};

export const DropDown = ({ label, value, options, onChange, required = false }: DropdownProps) => {
  const dropdownRef = React.useRef<HTMLDivElement>(null);
  const [isDropdownOpen, setIsDropdownOpen] = React.useState<boolean>(false);

  React.useEffect(() => {
    const handleClickOutside = (event: MouseEvent) => {
      if (dropdownRef.current && !dropdownRef.current.contains(event.target as Node)) {
        setIsDropdownOpen(false);
      }
    };
    if (isDropdownOpen) document.addEventListener('mousedown', handleClickOutside);
    return () => {
      document.removeEventListener('mousedown', handleClickOutside);
    };
  }, [isDropdownOpen]);

  return (
    <div className="w-full">
      {label && (
        <label className="label">
          <span
            className={`text-sm font-medium text-secondary-700 ${required && "after:content-['*'] after:ml-0.5 after:text-danger-500"}`}
          >
            {label}
          </span>
        </label>
      )}
      <DropdownContainer className="w-full" ref={dropdownRef}>
        <DropdownButton
          title={value}
          onClick={(e) => {
            e.stopPropagation();
            e.preventDefault();
            setIsDropdownOpen(!isDropdownOpen);
          }}
        />
        {isDropdownOpen && (
          <DropdownItemContainer align="left-0">
            {options &&
              options.map((item: any, index: number) => (
                <DropdownItem
                  key={index}
                  value={item.toString()}
                  onClick={() => {
                    const parsedValue = typeof item === 'number' ? item.toString() : item;
                    if (onChange) {
                      onChange(parsedValue);
                    }
                    setIsDropdownOpen(false);
                  }}
                />
              ))}
          </DropdownItemContainer>
        )}
      </DropdownContainer>
    </div>
  );
};

export default Input;
