import {
  useAuth,
  useAuthorizationCache,
  useAppDispatch,
  useSessionCache,
  useQuerybuilderGraph,
  useQuerybuilderHash,
  useML,
  useMLEnabledHash,
  useConfig,
  useQuerybuilderSettings,
  readInSchemaFromBackend,
  assignNewGraphQueryResult,
  setQuerybuilderNodes,
  resetGraphQueryResults,
  useQuerybuilder,
} from '@graphpolaris/shared/lib/data-access';
import { WebSocketHandler } from '@graphpolaris/shared/lib/data-access/socket';
import Broker from '@graphpolaris/shared/lib/data-access/socket/broker';
import { addError, addInfo } from '@graphpolaris/shared/lib/data-access/store/configSlice';
import { GraphQueryResultFromBackendPayload, queryingBackend } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { allMLTypes, LinkPredictionInstance, setMLResult } from '@graphpolaris/shared/lib/data-access/store/mlSlice';
import { QueryBuilderState } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { QueryMultiGraph, Query2BackendQuery } from '@graphpolaris/shared/lib/querybuilder';
import { SchemaFromBackend } from '@graphpolaris/shared/lib/schema';
import { useRef, useState, useEffect } from 'react';
import { DatabaseInfo, DatabaseStatus, SaveStateI, TestDatabaseConnectionResponse, wsGetStates } from './wsState';
import { wsSchemaRequest } from './wsSchema';
import { addSaveState, testedSaveState, updateSaveStateList } from '../store/sessionSlice';

export const EventBus = (props: { onRunQuery: Function; onAuthorized: Function }) => {
  const { login } = useAuth();
  const auth = useAuthorizationCache();
  const dispatch = useAppDispatch();
  const session = useSessionCache();
  const queryHash = useQuerybuilderHash();
  const queryBuilder = useQuerybuilder();
  const mlHash = useMLEnabledHash();
  const queryBuilderSettings = useQuerybuilderSettings();

  useEffect(() => {
    // Default
    Broker.instance().subscribe((data: SchemaFromBackend) => {
      dispatch(readInSchemaFromBackend(data));
      dispatch(addInfo('Schema graph updated'));
    }, 'schema_result');
    Broker.instance().subscribe((data: GraphQueryResultFromBackendPayload) => {
      dispatch(assignNewGraphQueryResult(data));
      dispatch(addInfo('Query Executed!'));
    }, 'query_result');
    // Broker.instance().subscribe((data: QueryBuilderState) => dispatch(setQuerybuilderNodes(data)), 'query_builder_state');
    allMLTypes.forEach((mlType) => {
      Broker.instance().subscribe((data: LinkPredictionInstance[]) => dispatch(setMLResult({ type: mlType, result: data })), mlType);
    });

    Broker.instance().subscribe((data: SaveStateI[]) => {
      dispatch(updateSaveStateList(data));
    }, 'save_states');
    Broker.instance().subscribe((data: any) => {}, 'save_state_status');
    Broker.instance().subscribe((data: SaveStateI) => {
      dispatch(addSaveState(data));
    }, 'save_state');
    Broker.instance().subscribe((data: SaveStateI) => {
      wsGetStates();
    }, 'delete_save_state');
    Broker.instance().subscribe((response: TestDatabaseConnectionResponse) => {
      if (response && response.status === 'success') dispatch(testedSaveState(response.saveStateID));
    }, 'tested_connection');

    login();

    return () => {
      Broker.instance().unSubscribeAll('schema_result');
      Broker.instance().unSubscribeAll('query_result');

      Broker.instance().unSubscribeAll('save_states');
      Broker.instance().unSubscribeAll('save_state');
      Broker.instance().unSubscribeAll('save_state_status');
      Broker.instance().unSubscribeAll('delete_save_state');
      Broker.instance().unSubscribeAll('tested_connection');
      // Broker.instance().unSubscribeAll('query_builder_state');
      allMLTypes.forEach((mlType) => {
        Broker.instance().unSubscribeAll(mlType);
      });
    };
  }, []);

  useEffect(() => {
    // New active database
    if (session.currentSaveState) {
      wsSchemaRequest(session.currentSaveState);
    }
  }, [session.currentSaveState]);

  useEffect(() => {
    // Newly (un)authorized
    if (auth.authorized && auth.jwt) {
      props.onAuthorized();
      WebSocketHandler.instance()
        .useAuth(auth)
        .connect(() => {
          wsGetStates();
          // WebSocketHandler.instance().sendMessage({ //TODO!!
          //   sessionID: auth?.sessionID || '',
          //   key: 'broadcastState',
          //   body: { type: 'subscribe', status: '', value: {} },
          // });
        });
    } else {
      // dispatch(logout());
    }
  }, [auth]);

  useEffect(() => {
    if (!queryBuilder.ignoreReactivity) {
      props.onRunQuery();
      // WebSocketHandler.instance().sendMessage({ //TODO!!
      //   sessionID: auth?.sessionID || '',
      //   key: 'broadcastState',
      //   body: { type: 'query_builder_state', status: '', value: queryBuilder },
      // });
    }
  }, [queryHash, mlHash, queryBuilderSettings]);

  return <div className="hide"></div>;
};
