import React, { useState, useEffect, useMemo } from 'react';
import * as d3 from 'd3';
import Pagination from '@graphpolaris/shared/lib/components/pagination';
import Icon from '@graphpolaris/shared/lib/components/icon';

import BarPlot from '@graphpolaris/shared/lib/components/charts/barplot';

import { NodeAttributes } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { SchemaAttributeTypes } from '@graphpolaris/shared/lib/schema';
import styles from './table.module.scss';

export type AugmentedNodeAttributes = { attribute: NodeAttributes; type: Record<string, SchemaAttributeTypes> };

type TableProps = {
  data: AugmentedNodeAttributes[];
  itemsPerPage: number;
  showBarPlot: boolean;
};
type Data2RenderI = {
  name: string;
  typeData: SchemaAttributeTypes;
  data: { category: string; count: number }[];
  numElements: number;
  showBarPlot?: boolean;
};

export const Table = ({ data, itemsPerPage, showBarPlot }: TableProps) => {
  const maxUniqueValues = 69;
  const barPlotNumBins = 10;

  const [sortedData, setSortedData] = useState<AugmentedNodeAttributes[]>(data);
  const [sortOrder, setSortOrder] = useState<'asc' | 'desc'>('asc');
  const [sortColumn, setSortColumn] = useState<string | null>(null);
  const [currentPage, setCurrentPage] = useState<{
    page: number;
    startIndex: number;
    endIndex: number;
    currentData: AugmentedNodeAttributes[];
  } | null>(null);
  const [data2Render, setData2Render] = useState<Data2RenderI[]>([]);
  const dataColumns = useMemo(() => Object.keys(data[0].attribute), [data]);
  const totalPages = Math.ceil(sortedData.length / itemsPerPage);

  useEffect(() => {
    if (sortColumn !== null) {
      const sorted = [...data].sort((a, b) => {
        if (sortOrder === 'asc') {
          return (a.attribute as any)[sortColumn] < (b.attribute as any)[sortColumn] ? -1 : 1;
        } else {
          return (a.attribute as any)[sortColumn] > (b.attribute as any)[sortColumn] ? -1 : 1;
        }
      });
      setSortedData(sorted);
    }
  }, [sortOrder, data, sortColumn]);

  useEffect(() => {
    onPageChange(1); // Reset to the first page when sorting or itemsPerPage changes
  }, [sortColumn, sortOrder, itemsPerPage]);

  const onPageChange = (page: number) => {
    const startIndex = (page - 1) * itemsPerPage;
    const endIndex = startIndex + itemsPerPage;
    const currentData = sortedData.slice(startIndex, endIndex);

    setCurrentPage({
      page: page,
      startIndex: startIndex,
      endIndex: endIndex,
      currentData: currentData,
    });
  };

  const toggleSort = (column: string) => {
    if (sortColumn === column) {
      if (sortOrder === 'asc') {
        setSortOrder('desc');
        setSortedData(sortedData.reverse()); // Reverse the sorted data
      } else if (sortOrder === 'desc') {
        setSortColumn(null);
        setSortOrder('asc');
        setSortedData(data); // Reset to the original order
      }
    } else {
      setSortColumn(column);
      setSortOrder('asc');
      const sorted = [...data].sort((a, b) => {
        return (a.attribute as any)[column] < (b.attribute as any)[column] ? -1 : 1;
      });
      setSortedData(sorted);
    }
  };

  // Barplot on headers data preparation

  // Data structure to feed the keys-barplot ( name data2Render )
  // Three options:
  // 1) numeric->data to histogram barplot
  // 2) string with unique values < maxUniqueValues -> data binned barplot
  // 3) string with unique values > maxUniqueValues -> show text unique values

  useEffect(() => {
    if (!currentPage || currentPage?.currentData?.length <= 0) return;

    let categoryCounts = [];
    const firstRowData = data[0];

    console.log('First row: ', firstRowData);

    let _data2Render = Object.keys(firstRowData.attribute).map((dataColumn: string, i) => {
      const newData2Render: Data2RenderI = {
        name: dataColumn,
        typeData: firstRowData.type[dataColumn] || 'string',
        data: [],
        numElements: 0,
      };

      if (
        firstRowData.type[dataColumn] === 'string' ||
        firstRowData.type[dataColumn] === 'date' ||
        firstRowData.type[dataColumn] === 'duration' ||
        firstRowData.type[dataColumn] === 'datetime' ||
        firstRowData.type[dataColumn] === 'time'
      ) {
        const groupedData = d3.group(data, (d) => d.attribute[dataColumn]);
        categoryCounts = Array.from(groupedData, ([category, items]) => ({
          category: category as string,
          count: items.length,
        }));

        if (categoryCounts.length > maxUniqueValues || categoryCounts.length <= 1 || !categoryCounts.some((c) => c.count > 1)) {
          newData2Render.numElements = categoryCounts.length;
          newData2Render.showBarPlot = false;
        } else {
          newData2Render.numElements = categoryCounts.length;
          newData2Render.data = categoryCounts;
          newData2Render.showBarPlot = true;
        }
      } else if (firstRowData.type[dataColumn] === 'bool') {
        const groupedData = d3.group(data, (d) => d.attribute[dataColumn]);

        categoryCounts = Array.from(groupedData, ([category, items]) => ({
          category: category as string,
          count: items.length,
        }));

        newData2Render.numElements = categoryCounts.length;
        newData2Render.data = categoryCounts;
        newData2Render.showBarPlot = true;

        // number: float and int
      } else if (firstRowData.type[dataColumn] === 'bool') {
        const groupedData = d3.group(data, (d) => d.attribute[dataColumn]);

        categoryCounts = Array.from(groupedData, ([category, items]) => ({
          category: category as string,
          count: items.length,
        }));

        newData2Render.numElements = categoryCounts.length;
        newData2Render.data = categoryCounts;
        newData2Render.showBarPlot = true;

        // number: float and int
      } else if (firstRowData.type[dataColumn] === 'int' || firstRowData.type[dataColumn] === 'float') {
        categoryCounts = data.map((obj) => ({
          category: 'placeholder', // add something
          count: obj.attribute[dataColumn] as number, // fill values of data
        }));

        newData2Render.numElements = categoryCounts.length;
        newData2Render.data = categoryCounts;
        newData2Render.showBarPlot = true;
      } else {
        // there is also array type, when considering labels
        const groupedData = d3.group(data, (d) => (d.attribute[dataColumn] as any)?.[0]);

        categoryCounts = Array.from(groupedData, ([category, items]) => ({
          category: category as string,
          count: items.length,
        }));

        newData2Render.numElements = categoryCounts.length;
        newData2Render.showBarPlot = false;
      }
      return newData2Render;
    });

    console.log(_data2Render);

    setData2Render(_data2Render);
  }, [currentPage, data, sortedData]);

  return (
    <>
      {currentPage && currentPage?.currentData?.length > 0 && data2Render?.length > 0 && (
        <>
          <div className={styles['table-container']}>
            <table className={`${styles.table} bg-secondary-100`}>
              <thead className="thead">
                <tr>
                  {dataColumns.map((item, index) => (
                    <th
                      className="border-light group hover:bg-secondary-300 bg-secondary-200 text-left overflow-x-hidden truncate capitalize cursor-pointer"
                      key={index + item}
                      onClick={() => toggleSort(item)}
                    >
                      <div className="flex flex-row max-w-full gap-1 justify-between">
                        <span className="shrink overflow-hidden text-ellipsis">{item}</span>
                        <div
                          className={
                            sortColumn === item ? 'opacity-100 text-primary' : 'opacity-0 group-hover:opacity-100 text-secondary-400'
                          }
                        >
                          <Icon
                            name={sortColumn === item ? (sortOrder === 'asc' ? 'ArrowDownward' : 'ArrowUpward') : 'Sort'}
                            size={20}
                            className={`${sortColumn === item && sortOrder ? 'text-secondary-800' : 'text-secondary-500'}`}
                          />
                        </div>
                      </div>
                    </th>
                  ))}
                </tr>
                <tr>
                  {dataColumns.map((item, index) => (
                    <th className="border-light bg-light">
                      <div className="th" key={index + item}>
                        <div className="h-full w-full overflow-hidden">
                          {data2Render[index].showBarPlot &&
                          (data2Render[index]?.typeData === 'int' || data2Render[index]?.typeData === 'float') ? (
                            <BarPlot typeBarPlot="numerical" numBins={barPlotNumBins} data={data2Render[index].data} />
                          ) : !data2Render[index]?.typeData || !data2Render[index].showBarPlot ? (
                            <div className="font-normal mx-auto flex flex-row items-start justify-center w-full gap-1 text-center text-secondary-700">
                              <span className="text-2xs overflow-x-hidden truncate">Unique values:</span>
                              <span className="text-xs shrink-0">{data2Render[index]?.numElements}</span>
                            </div>
                          ) : (
                            <BarPlot typeBarPlot="categorical" numBins={barPlotNumBins} data={data2Render[index].data} />
                          )}
                        </div>
                      </div>
                    </th>
                  ))}
                </tr>
              </thead>
              <tbody>
                {currentPage.currentData.map((item, rowIndex) => (
                  <tr key={rowIndex}>
                    {dataColumns.map((col, colIndex) => (
                      <td className="border-light" data-datatype={item.type[col]} key={colIndex}>
                        {item.attribute[col] !== undefined ? (item.attribute[col] as any).toString() : item.attribute[col]}
                      </td>
                    ))}
                  </tr>
                ))}
              </tbody>
            </table>
          </div>
          <Pagination
            currentPage={currentPage.page}
            totalPages={totalPages}
            onPageChange={onPageChange}
            itemsPerPageInput={itemsPerPage}
            numItemsArrayReal={currentPage.startIndex + currentPage.currentData.length}
            totalItems={sortedData.length}
          />
        </>
      )}
    </>
  );
};

export default Table;
